from asyncio import BaseProtocol
from enum import Enum
from xml.etree import ElementTree as ET
from xml.sax import ContentHandler

from loguru import logger

from pyjabber.metadata import host
from pyjabber.stream import Stream
from pyjabber.stream.StanzaHandler import StanzaHandler
from pyjabber.stream.StreamHandler import Signal, StreamHandler
from pyjabber.utils import ClarkNotation as CN


class XMLParser(ContentHandler):
    """
        Manages the stream data and process the XML objects.
        Inheriting from sax.ContentHandler

        :param buffer: Transport instance of the connected client. Used to send replays
        :param starttls: Coroutine launched when server and client start the connection upgrade process to TLS
    """

    def __init__(self, buffer, starttls):
        super().__init__()

        self._host = host.get()
        self._buffer = buffer

        self._state = self.StreamState.CONNECTED
        self._stanzaHandler = None
        self._stack = []
        self._streamHandler = StreamHandler(self._buffer, starttls)

    class StreamState(Enum):
        """
        Stream connection states.
        """
        CONNECTED = 0
        READY = 1

    @property
    def buffer(self) -> BaseProtocol:
        return self._buffer

    @buffer.setter
    def buffer(self, value: BaseProtocol):
        self._buffer = value
        self._streamHandler.buffer = value

    def startElementNS(self, name, qname, attrs):
        # logger.trace(f"Start element from <{hex(id(self._buffer))}>: {name}")

        if self._stack:  # "<stream:stream>" tag already present in the data stack
            elem = ET.Element(
                CN.clarkFromTuple(name),
                attrib={
                    CN.clarkFromTuple(key): item for key,
                    item in dict(attrs).items()})
            self._stack.append(elem)

        elif name[1] == "stream" and name[0] == "http://etherx.jabber.org/streams":
            self._buffer.write(Stream.responseStream(attrs))

            elem = ET.Element(
                CN.clarkFromTuple(name),
                attrib={
                    CN.clarkFromTuple(key): item for key,
                    item in dict(attrs).items()})

            self._stack.append(elem)
            self._streamHandler.handle_open_stream()

        else:
            raise Exception()

    def endElementNS(self, name, qname):
        # logger.trace(f"End element NS: {qname} : {name}")

        if "stream" in name:
            self._buffer.write(b'</stream:stream>')
            self._stack.clear()
            return

        if not self._stack:
            raise Exception()

        elem = self._stack.pop()

        if elem.tag != CN.clarkFromTuple(name):
            # INVALID STANZA/MESSAGE
            raise Exception()

        if self._stack[-1].tag != '{http://etherx.jabber.org/streams}stream':
            self._stack[-1].append(elem)

        else:
            if self._state == self.StreamState.READY:  # Ready to process stanzas
                self._stanzaHandler.feed(elem)
            else:
                signal = self._streamHandler.handle_open_stream(elem)
                if signal == Signal.RESET and "stream" in self._stack[-1].tag:
                    self._stack.clear()
                elif signal == Signal.DONE:
                    self._stanzaHandler = StanzaHandler(self._host, self._buffer)
                    self._state = self.StreamState.READY

    def characters(self, content: str) -> None:
        if not self._stack:
            raise Exception()

        elem = self._stack[-1]
        if len(elem) != 0:
            child = elem[-1]
            child.tail = (child.tail or '') + content

        else:
            elem.text = (elem.text or '') + content



import contextvars

host = contextvars.ContextVar("host")
config_path = contextvars.ContextVar("config_path")
root_path = contextvars.ContextVar("root_path")
database_path = contextvars.ContextVar("database_path")
database_in_memory = contextvars.ContextVar("database_in_memory")



from unittest.mock import MagicMock, patch
from xml.etree.ElementTree import Element
import xml.etree.ElementTree as ET
import pickle
from pyjabber.utils import ClarkNotation as CN
import os
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stream.StanzaHandler import StanzaHandler

def setUp():
    mock_buffer = MagicMock()
    mock_buffer.get_extra_info.return_value = '127.0.0.1'
    mock_host = "domain.es"

    with patch('pyjabber.network.ConnectionManager.ConnectionManager') as MockConnectionsManager, \
         patch('pyjabber.stream.StanzaHandler.PluginManager') as MockPluginManager, \
         patch('pyjabber.stream.StanzaHandler.Presence') as MockPresence:

        mock_plugin = MagicMock()
        MockPluginManager.return_value = mock_plugin

        mock_presence = MagicMock()
        MockPresence.return_value = mock_presence

        mock_connections = MockConnectionsManager.return_value
        mock_connections.get_jid.return_value = 'user@domain.com'
        MockConnectionsManager.return_value = mock_connections

        handler = StanzaHandler(mock_host, mock_buffer)
        handler._host = mock_host
        handler._functions = {
            "{jabber:client}iq": MagicMock(),
            "{jabber:client}message": MagicMock(),
            "{jabber:client}presence": MagicMock()
        }
        handler._connection = mock_connections
        handler._presenceManager = mock_presence

    return handler, mock_buffer, mock_connections, mock_presence

def test_feed_valid_element():
    handler, mock_buffer, mock_connections, mock_presence = setUp()
    element = Element('iq', attrib={"to": "localhost", "type": "get"})
    element.tag = "{jabber:client}iq"
    child = Element('query')
    element.append(child)

    handler.feed(element)

    handler._functions[element.tag].assert_called_once_with(element)


@patch('pyjabber.stanzas.error.StanzaError.feature_not_implemented')
def test_feed_function_key_error(mock_feature_not_implemented):
    handler, mock_buffer, mock_connections, mock_presence = setUp()
    element = Element('unknown', attrib={"to": "localhost"})
    element.tag = "{jabber:client}unknown"
    child = Element('{jabber:client}query')
    element.append(child)

    # Simula que el esquema es válido
    try:
        handler.feed(element)
    except Exception:
        # Verifica que la excepción se lanzó
        pass
    else:
        raise AssertionError("Exception not raised")

    # No debería llegar aquí porque se debe lanzar una excepción
    mock_feature_not_implemented.assert_not_called()
    mock_buffer.write.assert_not_called()


def test_handleIQ():
    handler, mock_buffer, mock_connections, mock_presence = setUp()
    element = Element('iq', attrib={"to": "localhost", "type": "get"})
    child = Element('{jabber:client}query')  # Aseguramos que tenga el namespace adecuado
    element.append(child)

    expected_response = SE.service_unavaliable()

    with patch('pyjabber.stream.StanzaHandler.PluginManager') as MockPluginManager:
        MockPluginManager.feed.return_value = expected_response  # Simulamos la respuesta esperada
        handler._pluginManager = MockPluginManager
        handler.handle_iq(element)
        mock_buffer.write.assert_called_once_with(expected_response)


def test_handleMsg():
    handler, mock_buffer, mock_connections, mock_presence = setUp()
    element = Element('message', attrib={"to": "user@domain.com"})
    element.tag = "{jabber:client}message"

    mock_server_buffer = MagicMock()
    mock_connections.get_server_buffer.return_value = (None, mock_server_buffer)

    handler.handle_msg(element)

    mock_server_buffer.write.assert_called_once_with(ET.tostring(element))


def test_handleMsg_remote_user():
    handler, mock_buffer, mock_connections, mock_presence = setUp()
    element = Element('message', attrib={"to": "user@remote.com"})
    element.tag = "{jabber:client}message"
    mock_connections.get_server_buffer.return_value = None
    handler.handle_msg(element)
    mock_queue_message.enqueue.assert_called_once_with('remote.com', ET.tostring(element))

def test_handlePre():
    handler, mock_buffer, mock_connections, mock_presence = setUp()
    element = Element('presence', attrib={"to": "localhost"})
    mock_presence.feed.return_value = 'response'
    handler.handle_pre(element)
    mock_buffer.write.assert_called_once_with('response')



import asyncio
from unittest.mock import MagicMock, patch, AsyncMock
import pytest
from xml import sax
import ssl
from pyjabber.network.XMLProtocol import XMLProtocol, FILE_AUTH
import os


@patch('pyjabber.network.XMLProtocol.sax.make_parser')
def test_connection_made(mock_make_parser):
    mock_transport = MagicMock()
    mock_parser = MagicMock()
    mock_make_parser.return_value = mock_parser
    mock_connections_manager = MagicMock()
    mock_handler = MagicMock()

    namespace = "jabber:client"
    host = "domain.com"
    cert_path = None
    connection_timeout = 30

    with patch('pyjabber.network.XMLProtocol.XMLParser') as mock_handler:
        with patch('pyjabber.stream.StreamHandler.host') as mock_host_stream:
            mock_handler.return_value = MagicMock()
            mock_host_stream.get.return_value = 'localhost'

            protocol = XMLProtocol(
                host=host,
                namespace=namespace,
                connection_timeout=connection_timeout,
                cert_path=cert_path,
            )
            protocol._connection_manager = mock_connections_manager
            protocol.connection_made(mock_transport)

        mock_transport.get_extra_info.assert_called_with('peername')
        mock_make_parser.assert_called_once()
        mock_parser.setFeature.assert_any_call(sax.handler.feature_namespaces, True)
        mock_parser.setFeature.assert_any_call(sax.handler.feature_external_ges, False)
        mock_parser.setContentHandler.assert_called_once_with(mock_handler.return_value)
        mock_connections_manager.connection.assert_called_once_with(mock_transport.get_extra_info(), mock_transport)


@patch('pyjabber.network.XMLProtocol.StreamAlivenessMonitor')
def test_connection_made_with_timeout(mock_monitor):
    mock_transport = MagicMock()
    mock_parser = MagicMock()

    namespace = "jabber:client"
    host = MagicMock()
    cert_path = None
    connection_timeout = 30

    with patch('pyjabber.network.XMLProtocol.sax.make_parser', return_value=mock_parser):
        with patch('pyjabber.network.XMLProtocol.XMLParser') as mock_handler:
            mock_handler.return_value = MagicMock()
            protocol = XMLProtocol(
                host=host,
                namespace=namespace,
                cert_path=cert_path,
                connection_timeout=connection_timeout,
            )
            protocol.connection_made(mock_transport)

    mock_monitor.assert_called_once_with(timeout=30, callback=protocol.connection_timeout)
    mock_transport.get_extra_info.assert_called_with('peername')


def test_connection_lost():
    namespace = "jabber:client"
    host = MagicMock()
    cert_path = None
    connection_timeout = 30

    protocol = XMLProtocol(
        host=host,
        namespace=namespace,
        connection_timeout=connection_timeout,
        cert_path=cert_path,
    )
    mock_transport = MagicMock()
    protocol._transport = mock_transport

    protocol.connection_lost(None)
    assert protocol._transport is None
    assert protocol._xml_parser is None


def test_data_received():
    namespace = "jabber:client"
    host = MagicMock()
    cert_path = None
    connection_timeout = 30

    protocol = XMLProtocol(
        host=host,
        namespace=namespace,
        connection_timeout=connection_timeout,
        cert_path=cert_path,
    )
    mock_transport = MagicMock()
    protocol._transport = mock_transport
    mock_parser = MagicMock()
    protocol._xml_parser = mock_parser
    mock_monitor = MagicMock()
    protocol._timeout_monitor = mock_monitor

    data = b"<?xml version=\'1.0\'?><stream:stream>"
    protocol.data_received(data)
    mock_monitor.reset.assert_called_once()
    mock_parser.feed.assert_called_once_with(b"<stream:stream>")


def test_eof_received():
    namespace = "jabber:client"
    host = MagicMock()
    cert_path = None
    connection_timeout = 30

    with patch('pyjabber.network.XMLParser.host') as mock_host:
        with patch('pyjabber.stream.StreamHandler.host') as mock_host_stream:
            mock_host.get.return_value = 'localhost'
            mock_host_stream.get.return_value = 'localhost'
            protocol = XMLProtocol(
                host=host,
                namespace=namespace,
                connection_timeout=connection_timeout,
                cert_path=cert_path,
            )
            mock_transport = MagicMock()
            protocol.connection_made(mock_transport)
            mock_connections = MagicMock()
            protocol._connection_manager = mock_connections

            protocol.eof_received()
            mock_transport.get_extra_info.assert_called_with('peername')
            mock_connections.disconnection.assert_called_once_with(mock_transport.get_extra_info.return_value)

    # Verificar que el transporte y el analizador XML se mantengan en su estado actual
    assert protocol._transport == mock_transport
    assert protocol._xml_parser is not None  # El parser debe estar inicializado en connection_made



def test_connection_timeout():
    namespace = "jabber:client"
    host = MagicMock()
    cert_path = None
    connection_timeout = 30

    protocol = XMLProtocol(
        host=host,
        namespace=namespace,
        connection_timeout=connection_timeout,
        cert_path=cert_path,
    )
    mock_transport = MagicMock()
    protocol._transport = mock_transport

    protocol.connection_timeout()
    mock_transport.write.assert_called_once_with(b"<connection-timeout/>")
    mock_transport.close.assert_called_once()
    assert protocol._transport is None
    assert protocol._xml_parser is None


@patch('pyjabber.network.XMLProtocol.asyncio.get_running_loop')
@patch('pyjabber.network.XMLProtocol.ssl.create_default_context')
def test_enable_tls(mock_create_default_context, mock_get_running_loop):
    namespace = "jabber:client"
    host = MagicMock()
    cert_path = None
    connection_timeout = 30

    protocol = XMLProtocol(
        host=host,
        namespace=namespace,
        connection_timeout=connection_timeout,
        cert_path=cert_path,
    )
    mock_transport = MagicMock()
    protocol._transport = mock_transport
    mock_loop = MagicMock()
    mock_get_running_loop.return_value = mock_loop

    # Crear un contexto SSL real para evitar el error isinstance
    real_ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    mock_create_default_context.return_value = real_ssl_context

    # Mockear el parser para evitar el error 'NoneType'
    mock_parser = MagicMock()
    protocol._xml_parser = mock_parser

    # Asegurarse de que start_tls es un AsyncMock
    mock_loop.start_tls = AsyncMock()

    async def run_enable_tls():
        await protocol.enable_tls()
        mock_create_default_context.assert_called_once_with(ssl.Purpose.CLIENT_AUTH)
        real_ssl_context.load_cert_chain.assert_called_once_with(
            certfile=os.path.join(FILE_AUTH, 'certs', 'localhost.pem'),
            keyfile=os.path.join(FILE_AUTH, 'certs', 'localhost-key.pem')
        )
        mock_loop.start_tls.assert_called_once_with(
            transport=mock_transport,
            protocol=protocol,
            sslcontext=real_ssl_context,
            server_side=True
        )

    mock_loop.run_until_complete(run_enable_tls())




from uuid import uuid4
import pytest

from pyjabber.stream.JID import JID


def test_standard_jid():
    jid = JID(jid='demotest@host')

    assert jid.user == 'demotest'
    assert jid.domain == 'host'
    assert jid.resource is None


def test_user_and_domain_jid():
    jid = JID(user='demotest', domain='host')

    assert jid.user == 'demotest'
    assert jid.domain == 'host'
    assert jid.resource is None


def test_user_and_domain_and_resource_jid():
    resource = str(uuid4())
    jid = JID(user='demotest', domain='host', resource=resource)

    assert jid.user == 'demotest'
    assert jid.domain == 'host'
    assert jid.resource == resource


def test_to_string_jid():
    jid_string = f'demo@host/{str(uuid4())}'
    jid = JID(jid_string)

    assert jid_string == str(jid)


def test_bad_arguments():
    with pytest.raises(ValueError):
        JID(jid='demo@host', user='demo')
    with pytest.raises(ValueError):
        JID(jid='demo@host', domain='demo')
    with pytest.raises(ValueError):
        JID(jid='demo@host', resource='demo')
    with pytest.raises(ValueError):
        JID(jid='demo')



import asyncio
import os
import ssl
from xml import sax

from loguru import logger

from pyjabber.network.server.outcoming.XMLServerOutcomingParser import (
    XMLServerOutcomingParser,
)
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.XMLProtocol import XMLProtocol

FILE_AUTH = os.path.dirname(os.path.abspath(__file__))


class XMLServerOutcomingProtocol(XMLProtocol):
    """
    Protocol to manage the network connection between nodes in the XMPP network. Handles the transport layer.
    """

    def __init__(
            self,
            namespace,
            host,
            public_host,
            connection_timeout):

        super().__init__(
            namespace,
            host,
            connection_timeout,
            None)

        self._public_host = public_host

    def connection_made(self, transport):
        """
        Called when a client or another server opens a TCP connection to the server

        :param transport: The transport object for the connection
        :type transport: asyncio.Transport
        """
        if transport:
            self._transport = transport

            self._xml_parser = sax.make_parser()
            self._xml_parser.setFeature(sax.handler.feature_namespaces, True)
            self._xml_parser.setFeature(
                sax.handler.feature_external_ges, False)
            self._xml_parser.setContentHandler(
                XMLServerOutcomingParser(
                    self._transport,
                    self.task_tls,
                    self._host,
                    self._public_host,
                )
            )

            if self._connection_timeout:
                self._timeout_monitor = StreamAlivenessMonitor(
                    timeout=self._connection_timeout,
                    callback=self.connection_timeout
                )

            self._connection_manager.connection_server(
                self._transport.get_extra_info('peername'), self._host, self._transport)

            logger.info(
                f"Server connection to {self._transport.get_extra_info('peername')}")

        else:
            logger.error("Invalid transport")

    def eof_received(self):
        """
        Called when the client or another server sends an EOF
        """
        peer = self._transport.get_extra_info('peername')

        logger.debug(f"EOF received from {peer}")

        self._connection_manager.disconnection_server(peer)

        self._transport = None
        self._xml_parser = None

    ###########################################################################
    ###########################################################################
    ###########################################################################

    def task_tls(self):
        """
            Sync function to call the STARTTLS coroutine
        """
        loop = asyncio.get_running_loop()
        asyncio.ensure_future(self.enable_tls(loop), loop=loop)

    async def enable_tls(self, loop: asyncio.AbstractEventLoop):
        parser = self._xml_parser.getContentHandler()

        ssl_context = ssl.create_default_context()
        if not self._enable_tls1_3:
            ssl_context.options |= ssl.OP_NO_TLSv1_3

        new_transport = await loop.start_tls(
            transport=self._transport,
            protocol=self,
            sslcontext=ssl_context,
        )

        self._transport = new_transport
        parser.buffer = self._transport

        logger.debug("Done TLS")
